#include"core.h"  

using namespace std; 

Molecule::Molecule( string s ) {

name = s; 
site_count = 0; 

} 



Component::Component(string s) {
        
        name = s; 
	state = "free"; 
        

}


Cell::Cell() {
	count["R"] = 0; 
        count["L"] = 0; 

} 

void Cell::create_solution_compartment() {
	Compartment* c = new Compartment(); 
        c->name = "solution"; 
        solution = c; 
}; 


void Cell::create_membrane_compartment() {
	Compartment* c = new Compartment(); 
        c->name = "membrane"; 
        membrane = c; 
} 

void Cell::create_receptor_molecule(int n) {
         
	for (int i = 0; i < n; i++) {
    		Molecule* m = new Molecule("R"); 
    		++count[m->name]; 
    		m->id = count[m->name];  
    		m->partner = NULL; 
                m->create_site("Rsite", n_rec_site); 
                m->compartment = membrane; 
                molecule_list[m->name].push_back(m); 
	} 
} 



void Cell::create_ligand_molecule(int n) {
         
	for (int i = 0; i < n; i++) {
    		Molecule* m = new Molecule("L"); 
    		++count[m->name]; 
    		m->id = count[m->name];  
    		m->partner = NULL; 
                m->create_site("Lsite", n_lig_site); 
                m->compartment = solution; 
                m->n_site_bound = 0; 
                molecule_list[m->name].push_back(m); 
	} 
} 

void Cell::define_reactions() {
	Reaction* r = new Reaction(); 
        r->id = 0; 
        r->A = "Lsite_solution_free"; 
        r->B = "Rsite_membrane_free";  
        r->ratek = kf; 
        r->rate = 0; 
        reaction_list.push_back(r); 

        r = new Reaction(); 
        r->id = 1; 
        r->A = "Lsite_membrane_bound_1"; 
        r->B = "na"; 
        r->ratek = kr; 
        r->rate = 0; 
        reaction_list.push_back(r); 

        r = new Reaction(); 
        r->id = 2; 
        r->A = "Lsite_membrane_free"; 
        r->B = "Rsite_membrane_free"; 
        r->ratek = kx; 
        r->rate = 0; 
        reaction_list.push_back(r); 

        r = new Reaction(); 
        r->id = 3; 
        r->A = "Lsite_membrane_bound_2"; 
        r->B = "na"; 
        r->ratek = krx; 
        r->rate = 0; 
        reaction_list.push_back(r); 
} 


void Cell::update_reactant_list() {
        Molecule* m; 
        Component* c; 
        string s; 
        reactant_list.clear(); 

	vector<Molecule*> rlist = molecule_list["R"]; 
        for (vector<Molecule*>::const_iterator it = rlist.begin(); it != rlist.end(); it++) {
        	m = (*it); 
                vector<Component*> clist = m->site; 
                for (vector<Component*>::const_iterator jt = clist.begin(); jt != clist.end(); jt++) {
                        c = (*jt); 
                	s = c->name + "_" + m->compartment->name + "_" + c->state; 
                        reactant_list[s].push_back(c); 
                }  
        	
	} 


	rlist = molecule_list["L"]; 
        for (vector<Molecule*>::const_iterator it = rlist.begin(); it != rlist.end(); it++) {
        	m = (*it); 
                vector<Component*> clist = m->site; 
                for (vector<Component*>::const_iterator jt = clist.begin(); jt != clist.end(); jt++) {
                        c = (*jt); 
                        if (!(strcmp(m->compartment->name.c_str(),"membrane"))) {
                            if (m->n_site_bound == 1) 
                	        s = c->name + "_" + m->compartment->name + "_" + c->state + "_" + "1"; 
                            else 
                	        s = c->name + "_" + m->compartment->name + "_" + c->state + "_" + "2"; 

                            reactant_list[s].push_back(c); 

                            s = c->name + "_" + m->compartment->name + "_" + c->state; 
                            reactant_list[s].push_back(c); 
                        }
                        else { 
                           s = c->name + "_" + m->compartment->name + "_" + c->state; 
                           reactant_list[s].push_back(c); 
                        } 

                         
                }  
                
	} 


      
} 

void Cell::propagate_reaction() {
        double t = 0; 
        double a_tot = 0; 
        time = 0; 

        Reaction* rxn = NULL; 
        
	while(t < simulation_time) {
                
        	update_reactant_list(); 
                for(map<string,vector<Component*> >::const_iterator jt = reactant_list.begin(); jt != reactant_list.end(); jt++) {
                  //  cout << jt->first << "    " << (jt->second).size() << endl; 
                } 


                //getchar(); 

                compute_rate(); 
                a_tot = compute_sum(); 

                t += -log(RAN())/a_tot; 

                rxn = select_reaction(a_tot); 
                //cout << rxn->id << "     " << a_tot << endl; 
                execute_reaction(rxn->id); 

               for (map<string,vector<Component*> >::const_iterator it = reactant_list.begin(); it != reactant_list.end(); it++) {
                    // cout << it->first << "   " << endl; 
               } 

               if ( t - time > time_step) {
                    time = t; 
                    //save_species(); 
                    save_receptor_aggregate(); 
                    save_amount_ligand_bound(); 
                    cout << "Time " << "    " << time << endl; 
               } 
        } 

} 


void Cell::compute_rate() {
    for (map<string,vector<Component*> >::const_iterator it = reactant_list.begin(); it != reactant_list.end(); it++) {
          
    } 

    Reaction* rxn; 

    rxn = reaction_list[0]; 
    rxn->rate = (rxn->ratek)*(reactant_list[rxn->A].size())*(reactant_list[rxn->B].size());   

    rxn = reaction_list[1]; 
    rxn->rate = (rxn->ratek)*(reactant_list[rxn->A].size());   

    rxn = reaction_list[2]; 
    rxn->rate = (rxn->ratek)*(reactant_list[rxn->A].size())*(reactant_list[rxn->B].size());   

    rxn = reaction_list[3]; 
    rxn->rate = (rxn->ratek)*(reactant_list[rxn->A].size());   
} 


double Cell::compute_sum() {
    double sum = 0; 
    for (vector<Reaction*>::const_iterator it = reaction_list.begin(); it != reaction_list.end(); it++) {
        sum += (*it)->rate; 
    }
    
    return sum; 
} 


Reaction* Cell::select_reaction(double a_tot) {
    double a_1 = 0; 
    double a_ = RAN()*a_tot; 
    Reaction* rxn = NULL; 

    for (vector<Reaction*>::const_iterator it = reaction_list.begin(); it != reaction_list.end(); it++) {
        a_1 += (*it)->rate; 
        if (a_1 > a_) {
            rxn = (*it); 
            break; 
        }
    } 

    return rxn; 
}
         
         
        

void Cell::execute_reaction(int r) {
    Reaction* rxn = reaction_list[r]; 
    rxn = reaction_list[r];  
    string A, B; 
    Component* rA; 
    Component* rB; 
    int nA, nB; 

    if (r == 0) {
        A = rxn->A; 
        B = rxn->B; 
        nA = (int)(RAN()*reactant_list[A].size()); 
        nB = (int)(RAN()*reactant_list[B].size()); 
        
        rA = reactant_list[A][nA]; 
        rB = reactant_list[B][nB]; 

        rA->partner = rB; 
        rB->partner = rA; 

        rA->state = "bound"; 
        rA->molecule->compartment = membrane; 
        rB->state = "bound"; 
        ++(rA->molecule->n_site_bound); 
        ++(rB->molecule->n_site_bound); 
    } 
    else if (r == 1) {
        A = rxn->A; 
        nA = (int) (RAN()*reactant_list[A].size()); 
        
        rA = reactant_list[A][nA]; 
        rA->state = "free"; 
        rA->molecule->compartment = solution; 
        --(rA->molecule->n_site_bound); 
 
        rB = rA->partner; 
        rB->state = "free"; 
        --(rB->molecule->n_site_bound); 
 
        rB->partner = NULL; 
        rA->partner = NULL; 
    } 
    else if (r == 2) {
        A = rxn->A; 
        B = rxn->B; 
        nA = (int)(RAN()*reactant_list[A].size()); 
        nB = (int)(RAN()*reactant_list[B].size()); 
        
        rA = reactant_list[A][nA]; 
        rB = reactant_list[B][nB]; 

        Molecule* R = rB->molecule; 
        Molecule* L = rA->molecule; 

        if (reaction_decision_forward(R, L)) {  // Not an intracomplex binding 

            rA->partner = rB; 
            rB->partner = rA; 

            rA->state = "bound"; 
            rB->state = "bound"; 
            ++(rA->molecule->n_site_bound); 
            ++(rB->molecule->n_site_bound); 
        } 
    } 
        
    else {
        A = rxn->A; 
        nA = (int) (RAN()*reactant_list[A].size()); 
        
        rA = reactant_list[A][nA]; 
        rB = rA->partner; 
        
        Component* Rsite = rB; 
        Component* Lsite = rA; 

        bool x = true; 
        if (detailed_balance) 
            x = reaction_decision_reverse(Rsite,Lsite); 
 
        if (x) {
            rA->state = "free"; 
            --(rA->molecule->n_site_bound); 
 
            rB->state = "free"; 
            --(rB->molecule->n_site_bound); 

            rB->partner = NULL; 
            rA->partner = NULL; 
        } 
    } 


} 

void Cell::save_amount_ligand_bound() {
    //if (time < t_sample) return; 

    vector<Molecule*> mlist1 = molecule_list["L"]; 
    vector<Molecule*>::const_iterator it; 
    
    int sum = 0; 
 
    for (it = mlist1.begin(); it != mlist1.end(); it++) {
        Molecule* mol = (*it); 
        if (mol->compartment == membrane) {
            ++sum;          
        }
    } 

    pair<double,int> p; 
    p.first = time; 
    p.second = sum; 
    amount_ligand_bound.push_back(p); 
} 



void Cell::print_mean_amount_ligand_bound() {

    ofstream print_conc; 
    print_conc.open("mean_amount_ligand_bound.txt"); 

    int sum = 0; 
    double mean = 0; 

    vector<pair<double,int> >::const_iterator it; 
    pair<double,int> p; 

    int n_count = 0;  
    for (it = amount_ligand_bound.begin(); it != amount_ligand_bound.end(); it++) {
        p = (*it); 
       
        if (p.first > t_sample) {
            sum += p.second;
            ++n_count; 
        }   
    } 

    mean = (double) (1.0*sum/n_count); 

    print_conc << mean << endl; 
    print_conc.close(); 

}



void Cell::print_ligand_bound() {

    ofstream ligand_time_series;  
    ligand_time_series.open("ligand_bound_timeseries.txt"); 

    vector<pair<double,int> >::const_iterator it; 
    pair<double,int> p; 

    for (it = amount_ligand_bound.begin(); it != amount_ligand_bound.end(); it++) {
        p = (*it); 
        ligand_time_series << p.first << "            " << p.second << endl;    
    } 

    ligand_time_series.close(); 

}

void Cell::print_mean_aggregate_size() {
    ofstream print_mean_size; 
    print_mean_size.open("mean_aggregate_size.txt"); 
    int sum1 = 0; 
    int sum2 = 0; 
    double mean_size = 0;
 
    vector<pair<double,vector<vector<Molecule*> > > >::const_iterator it; 
    for (it = receptor_aggregate.begin(); it != receptor_aggregate.end(); it++) {
        pair<double, vector<vector<Molecule*> > > p = (*it); 
        vector<vector<Molecule*> > v = p.second; 
        vector<vector<Molecule*> >::const_iterator jt; 
        sum1 += v.size();  
        for (jt = v.begin(); jt != v.end(); jt++) {
            sum2 += (*jt).size();  
        } 
       
    } 

    mean_size = 1.0*sum2/sum1; 
    print_mean_size <<  mean_size << endl;
    print_mean_size.close();  
    
} 

void Cell::print_sol_gel_composition() {

    double fraction_in_gel = 0; 
    double number_of_complex = 0; 

    vector<pair<double,vector<vector<Molecule*> > > >::const_iterator it; 
    for (it = receptor_aggregate.begin(); it != receptor_aggregate.end(); it++) {
        pair<double, vector<vector<Molecule*> > > p = (*it); 
        vector<vector<Molecule*> > v = p.second; 
        vector<vector<Molecule*> >::const_iterator jt; 
        int sum = 0; 
        for (jt = v.begin(); jt != v.end(); jt++) {
            if ((*jt).size() > 1) { // If the complex size contains > 5% of total receptor it is in gel phase
                sum += 1; 
            } 
        } 
       
        number_of_complex += sum; 
        
    } 

    fraction_in_gel = number_of_complex/receptor_aggregate.size(); 
    ofstream receptor_in_gel; 
    receptor_in_gel.open("amount_receptor_in_gel.txt"); 
    receptor_in_gel << fraction_in_gel; 
    receptor_in_gel.close(); 

} 


/*int Cell::traverse_complex(Molecule* L, Molecule* R, vector<Molecule*>& v) {
    v.push_back(L); 
    cout << "-" << L->name << "(" << L->id << ")" << "-" ; 
    for (int i = 0; i < n_lig_site; i++ ) {
        Component* Lsite = L->site[i]; 
        if (Lsite->partner != NULL) {
            Component* Rsite = Lsite->partner; 
            Molecule* R1 = Rsite->molecule; 
            if (R1 != R) {
                cout << "-" << R1->name << "(" << R1->id << ")" << "-" ; 
                v.push_back(R1); 
                for (int j = 0; j < n_rec_site; j++) {
                    Component* Rsite1 = R1->site[j]; 
                    if (Rsite1->partner != NULL) {
                        Component* Lsite1 = Rsite1->partner; 
                        Molecule* L1 = Lsite1->molecule; 
                        if (L1 != L) {
                            traverse_complex(L1, R1, v); 
                        }
                    }
               }
          }

       } 

    } 

    return 0; 
} */ 

int Cell::traverse_complex(Molecule* R, Molecule* L, vector<Molecule*>& v) {
    v.push_back(R); 
    // cout << "-" << R->name << "(" << R->id << ")" << "-" ; 
    for (int i = 0; i < n_rec_site; i++ ) {
        Component* Rsite = R->site[i]; 
        if (Rsite->partner != NULL) {
            Component* Lsite = Rsite->partner; 
            Molecule* L1 = Lsite->molecule; 
            if (L1 != L) {
                // cout << "-" << L1->name << "(" << L1->id << ")" << "-" ; 
                v.push_back(L1); 
                for (int j = 0; j < n_lig_site; j++) {
                    Component* Lsite1 = L1->site[j]; 
                    if (Lsite1->partner != NULL) {
                        Component* Rsite1 = Lsite1->partner; 
                        Molecule* R1 = Rsite1->molecule; 
                        if (R1 != R) {
                            traverse_complex(R1, L1, v); 
                        }
                    }
               }
          }

       } 

    } 

    return 0; 
} 


int Cell::intra_complex(Molecule* R, Molecule* L, Molecule* L_look, int* intra) {
    for (int i = 0; i < n_rec_site; i++ ) {
        Component* Rsite = R->site[i]; 
        if (Rsite->partner != NULL) {
            Component* Lsite = Rsite->partner; 
            Molecule* L1 = Lsite->molecule; 

            if (L1 == L_look) {
                *intra = 1; 
                break; 
            }

            if (L1 != L) {
                for (int j = 0; j < n_lig_site; j++) {
                    Component* Lsite1 = L1->site[j]; 
                    if (Lsite1->partner != NULL) {
                        Component* Rsite1 = Lsite1->partner; 
                        Molecule* R1 = Rsite1->molecule; 
                        if (R1 != R) {
                            intra_complex(R1, L1, L_look, intra); 
                        }
                    }
               }
          }

       } 

    } 

    return 0; 
} 


int Cell::check_threshold_size(Molecule* R, Molecule* L, int* size) {
     if (*size > size_threshold) return 0; 
    //cout << "-" << R->name << "(" << R->id << ")" << "-"; 

    for (int i = 0; i < n_rec_site; i++ ) {
        Component* Rsite = R->site[i]; 
        if (Rsite->partner != NULL) {
            Component* Lsite = Rsite->partner; 
            Molecule* L1 = Lsite->molecule; 
            if (L1 != L) {
                //cout << "-" << L1->name << "(" << L1->id << ")" << "-"; 
                for (int j = 0; j < n_lig_site; j++) {
                    Component* Lsite1 = L1->site[j]; 
                    if (Lsite1->partner != NULL) {
                        Component* Rsite1 = Lsite1->partner; 
                        Molecule* R1 = Rsite1->molecule; 
                        if (R1 != R) {
                            ++(*size); 
                            check_threshold_size(R1, L1, size); 
                        }
                    }
               }
          }

       } 

    } 

    return 0; 
} 


int Cell::complex_size(Molecule* R, Molecule* L, vector<Molecule*>& v) {
    //cout << "-" << R->name << "(" << R->id << ")" << "-"; 

    v.push_back(R); 
    for (int i = 0; i < n_rec_site; i++ ) {
        Component* Rsite = R->site[i]; 
        if (Rsite->partner != NULL) {
            Component* Lsite = Rsite->partner; 
            Molecule* L1 = Lsite->molecule; 
            if (L1 != L) {
                //cout << "-" << L1->name << "(" << L1->id << ")" << "-"; 
                for (int j = 0; j < n_lig_site; j++) {
                    Component* Lsite1 = L1->site[j]; 
                    if (Lsite1->partner != NULL) {
                        Component* Rsite1 = Lsite1->partner; 
                        Molecule* R1 = Rsite1->molecule; 
                        if (R1 != R) {
                            complex_size(R1, L1, v); 
                        }
                    }
               }
          }

       } 

    } 

    return 0; 
} 


bool Cell::reaction_decision_forward(Molecule* R, Molecule* L) {
    int intra = 0; 

    intra_complex(R, NULL, L, &intra); 
    if (intra == 1) return false; // If intracomplex, reaction decision negative  

    int size1 = 1; 
    int size2 = 0; 

    if (intra == 0) { // If not intracomplex, check if both complexes are above the threshold size 

        Molecule* R1 = NULL;  
        for (int i = 0; i < n_lig_site; i++) { // Checking if the ligand is in complex with a receptor or not 
            Component* Lsite = L->site[i]; 
            if (Lsite->partner) {
                Component* Rsite = Lsite->partner; 
                R1 = Rsite->molecule; 
                size2 = 1; 
                break; 
            }  
        } 

        if (size2 == 1) { // If the ligand is in complex with any receptor 
            check_threshold_size(R,NULL,&size1); 
            check_threshold_size(R1,NULL,&size2); 
        } 

    } 


    if ((size1 > size_threshold) && (size2 > size_threshold)) { 
        //cout << size1 << "   " << size2 << endl; 
        //getchar(); 
        return false; // No reaction 
    } 

    return true; // Reaction 

} 


bool Cell::reaction_decision_reverse(Component* Rsite, Component* Lsite) {
    Molecule* R = Rsite->molecule; 
    Molecule* L = Lsite->molecule; 

    // Temporarily dissociate the two sites 
    Rsite->partner = NULL; 
    Lsite->partner = NULL; 

    Component* Rsite1 = NULL; 
    Molecule* R1 = NULL;  
    Component* Lsite1 = NULL; 

    int size1 = 1; 
    int size2 = 0; 
 
    for (int i = 0; i < n_lig_site; i++) {
        Lsite1 = L->site[i]; 
        if (Lsite1->partner) {
            Rsite1 = Lsite1->partner; 
            R1 = Rsite1->molecule; 
            size2 = 1; 
            break; 
        }
    } 
         
    if (size2 == 1) {
        //cout << endl << "First complex : "; 
        check_threshold_size(R,NULL,&size1); 
        //cout << endl << "Second complex : ";  
        check_threshold_size(R1,NULL,&size2); 
    } 

    // Re-conjugate the bond between the two sites 
    Rsite->partner = Lsite; 
    Lsite->partner = Rsite; 

    // cout << endl << "main complex " << endl; 
    int size3; 
    check_threshold_size(R,NULL,&size3); 

    if ((size1 > size_threshold) && (size2 > size_threshold)) { 
        //cout << endl << size1 << "       " << size2 << endl; 
        //getchar(); 
        return false; // No reaction if the both potential complexes are above the threshold size 
    } 

    return true; 

}  



void Cell::save_species() {

    vector<Molecule*> mlist1 = molecule_list["R"]; 

    vector< vector<Molecule*> > species_temp; 

    species_temp.clear(); 

    for (vector<Molecule*>::const_iterator it = mlist1.begin(); it != mlist1.end(); it++) {
        Molecule* R = (*it); 
        int k = 0; 
        vector<vector<Molecule*> >::iterator jt; 
   
        for (jt = species_temp.begin(); jt != species_temp.end(); jt++) {
            vector<Molecule*> spc = (*jt); 
            for (vector<Molecule*>::iterator kt = spc.begin(); kt != spc.end(); kt++) {
                if (R == (*kt)) {
                    k = 1; 
                    break;  
                }
            } 
        }

        if (k == 0) {
            vector<Molecule*> v; 
            traverse_complex(R,NULL,v); 
            species_temp.push_back(v); 
            //cout << endl << endl;  
        }
    }

    pair<double, vector<vector<Molecule*> > > p; 
    p.first = time; 
    p.second = species_temp; 

    species.push_back(p); 

} 

void Cell::save_receptor_aggregate() {
    if (time < t_sample) return; 

    vector<Molecule*> mlist1 = molecule_list["R"]; 

    vector< vector<Molecule*> > species_temp; 

    species_temp.clear(); 

    for (vector<Molecule*>::const_iterator it = mlist1.begin(); it != mlist1.end(); it++) {
        Molecule* R = (*it); 
        int k = 0; 
        vector<vector<Molecule*> >::iterator jt; 
   
        // Check if the receptor molecule has already enumerated in while evaluating a previous complex 
        for (jt = species_temp.begin(); jt != species_temp.end(); jt++) {
            vector<Molecule*> spc = (*jt); 
            for (vector<Molecule*>::iterator kt = spc.begin(); kt != spc.end(); kt++) {
                if (R == (*kt)) {
                    k = 1; 
                    break;  
                }
            } 
        }

        if (k == 0) {
            vector<Molecule*> v; 
            complex_size(R,NULL,v); 
            species_temp.push_back(v); 
            //cout << endl << endl;  
        }

    } 

    pair<double, vector<vector<Molecule*> > > p; 
    p.first = time; 
    p.second = species_temp; 

    receptor_aggregate.push_back(p); 
}


void Molecule::create_site(string s, int n) {
    for (int i = 0; i < n; i++) {
        ++site_count; 
        Component* c = new Component(s); 
        c->id = site_count; 
        c->molecule = this; 
        c->partner = NULL; 
        site.push_back(c); 
    } 

}


Compartment::Compartment(string s) {
	name = s; 

} 

/*Long period (> 2 × 10^18) random number generator of L'Ecuyer with Bays-Durham shuffle
and added safeguards. Returns a uniform random deviate between 0.0 and 1.0 (exclusive of
the endpoint values). Call with idum a negative integer to initialize; thereafter, do not
alter idum between successive deviates in a sequence. RNMX should approximate the largest
floating value that is less than 1.*/
#define IM1 2147483563
#define IM2 2147483399
#define AM (1.0/IM1)
#define IMM1 (IM1-1)
#define IA1 40014
#define IA2 40692
#define IQ1 53668
#define IQ2 52774
#define IR1 12211
#define IR2 3791
#define NTAB 32
#define NDIV (1+IMM1/NTAB)
#define EPS 1.2e-7
#define RNMX (1.0-EPS)
long idum=-536445;
double RAN(){

    int j;
    long k;
    static long idum2=123456789;
    static long iy=0;
    static long iv[NTAB];
    float temp;
    if (idum <= 0){                           // Initialize.
        if (-(idum) < 1) idum=1;              // Be sure to prevent idum = 0.
        else idum = -(idum);
        idum2=(idum);
        for(j=NTAB+7;j>=0;j--){               // Load the shuffle table (after 8 warm-ups).
           k=(idum)/IQ1;
           idum=IA1*(idum-k*IQ1)-k*IR1;
           if (idum < 0) idum += IM1;
           if (j < NTAB) iv[j] = idum;
        }
        iy=iv[0];
    }
    k=(idum)/IQ1;                             // Start here when not initializing.
    idum=IA1*(idum-k*IQ1)-k*IR1;              // Compute idum=(IA1*idum) % IM1 without
    if (idum < 0) idum += IM1;                //     overflows by Schrage's method.
    k=idum2/IQ2;
    idum2=IA2*(idum2-k*IQ2)-k*IR2;            // Compute idum2=(IA2*idum) % IM2 likewise.
    if (idum2 < 0) idum2 += IM2;
    j=iy/NDIV;                                // Will be in the range 0..NTAB-1.
    iy=iv[j]-idum2;                           // Here idum is shuffled, idum and idum2 are
    iv[j] = idum;                             //     combined to generate output.
    if (iy < 1) iy += IMM1;
    if ((temp=AM*iy) > RNMX)
       return RNMX;                           // Because users don't expect endpoint values.
    else
       return temp;
}

