#ifndef core_H_
#define core_H_
#include<cstdlib>
#include<iostream>
#include<fstream> 
#include<vector>
#include<map>
#include<cmath>
#include<ctime>
#include<algorithm> 
#include<cstdio> 
#include<cstring>  
#include<string> 

#define n_lig 100

#define n_rec 100  
#define n_rec_site 3 
#define n_lig_site 3 
//#define kf 0.00000166   // Association of a ligand molecule from solution to a recepto 
#define kf 0.0083   // Association of a ligand molecule from solution to a recepto 
#define kr 0.001 // Dissociation of a ligand molecule from a receptor that will release the ligand into the solution  
#define kx 0.0083   // Membrane crosslinking 
#define krx 0.001 // Membrane dissociation of complexes  
#define size_threshold 1e6 // Both complexes need to be smaller or equal to this threshold size in order for their association to happen  
#define detailed_balance 1 // Setting 1 will prohibit dissociation if both new complexes are greater than threshold size; setting zero will disable this constraint  
#define simulation_time 500
#define time_step 0.5 
#define t_sample 450 // beginning of sample time of simulation data  


using namespace std; 
void read_parameters(); 

double RAN(); 
class Compartment; 
class Component; 
class Molecule; 
class Cell; 
class Reaction; 
class SiteList; 
class Class; 

class Cell {
public: 
        Cell();  
        vector<Compartment*> compartment_list; 
	map<string, vector<Molecule*> >molecule_list; 
        vector<Reaction*> reaction_list; 
        map<string,vector<Component*> >reactant_list; 
        map<string, int> count; 
        Compartment* solution; 
        Compartment* membrane; 
        void create_receptor_molecule(int n); 
        void create_ligand_molecule(int n); 
        void create_solution_compartment(); 
        void create_membrane_compartment(); 
        void define_reactions(); 
        void update_reactant_list(); 
        void propagate_reaction(); 
        void compute_rate(); 
        double compute_sum(); 
        void execute_reaction(int r); 
        //int traverse_complex(Molecule* L, Molecule* R, vector<Molecule*>& v); 
        int traverse_complex(Molecule* R, Molecule* L, vector<Molecule*>& v); 
        int intra_complex(Molecule* R, Molecule* L, Molecule* L_look, int* intra); 
        int check_threshold_size(Molecule* R, Molecule* L, int* size); 
        int complex_size(Molecule* R, Molecule* L, vector<Molecule*>& v); 
        bool reaction_decision_forward(Molecule* R, Molecule* L); 
        bool reaction_decision_reverse(Component* Rsite, Component* Lsite); 
        Reaction* select_reaction(double a_tot); 
        vector<pair<double, vector< vector<Molecule*> > > > species; 
        vector<pair<double, vector< vector<Molecule*> > > > receptor_aggregate; 
        void save_species(); 
        void save_receptor_aggregate();
        void print_mean_amount_ligand_bound();
        void print_ligand_bound();
        void print_sol_gel_composition(); 
        void print_mean_aggregate_size(); 
        double time; 
private: 
        void save_amount_ligand_bound(); 
        vector<pair<double,int> >  amount_ligand_bound; 

}; 

class Molecule {
public: 
	Molecule() { } 
        int id; 
	Molecule(string name); 
	string name; 
        vector<Component*> site; 
        Compartment* compartment; 
        string state; 
        int site_count; 
        Molecule* partner; 
        int n_site_bound; 
        void create_site(string s, int n); 
           
private: 
}; 


class Component {
public: 
	Component() { } 
        Component(string s); 
        int id; 
        string name; 
        Molecule* molecule; 
        string state; 
        Reaction* reaction;  
        Component* partner; 


private: 

}; 


class Reaction {
public: 
	Reaction() { }
        int id; 
        string key; 
        string A; 
        string B; 
        double ratek; 
        double rate; 
private: 

}; 

class Compartment {
public: 
        Compartment() { } 
	Compartment(string name); 
        string name; 

private: 

};



#endif 
